
/*Implemented commands 
 *IDN?
 *RST
 *CLS
 SYST:ERR? 
     :ERR:ALL? 
     :ERR:CODE:COUN? 
     :ERR:CODE:ALL? 
     :COMM:SER:BAUD {}                            //Changes the Baud Rate
     :INT:TRIP?                                   //Queries the state of the interlock (1 is tripped, aka a disconnected interlock)
 OUTP:DUT {0..255}                                //Sets the duty cycle of the internal 7.8kHz PWM. Only available in operating modes A, C, D
     :DUT?                                        //Queries the current duty cycle
     :CUR {0..255}                                //Sets the LED current, only available in operating modes B, C, D
     :CUR                                         //Queries the current LED current
     :MOD {A|B|C|D}                               //Sets the operating mode. For explanation of operation modes see function implementation of "change_mode"
     :MOD?                                        //Queries current operating mode
     :EN?                                         //Queries whether output is enabled or disabled
     :EN {0|1}                                    //Enabled/disables the output

 TEMP?                                            //Queries the temperature of the LED board
 TEMP:LIM {0..125}                                //Sets the temperature at which the overtemperature protection acts. Value is stored in EEPROM and is stored persistently
     :LIM?                                        //Queries the current temperature limit


*/


#include <avr/wdt.h>
#include <EEPROM.h>

#define ERROR_QUEUE_LENGTH  10                    //Length of the corresponding arrays
#define COLON_QUEUE_LENGTH  7
#define TEMPERATURE_LIM_ADDR 0                    //EEPROM-Adress for temperature limit value


/***Pin aliases***/
  int AIN_INT = 11;                               // PWM output for internal I_MAX value
  int AIN_INT_EN = 8;                             // Output for enabling forwarding the internal I_MAX value (controls gate of Q5)
  int AIN_EXT = A7;                               // ADC input for avaluating the analog value of the signal input
  int AIN_EXT_EN = 7;                             // Output for enabling forwarding the external I_MAX value (controls gate of Q3)
  int AIN_EXT_INTERR = 2;                         // Interrupt input for the signal input
  int DIP1 = 9;                                   // DIP switch, first switch
  int DIP2 = 10;                                  //   "         , second switch
  int INTERLOCK = 4;                              // Evaluation for the hardware interlock
  int STROBE = 3;                                 // PWM-output for strobing the output (pwm dimming)
  int TEMP_SENS = A6;                             // Input for the temperature sensor

/***Global variables***/

  volatile unsigned char DUTYCYCLE_INV = 255;
  volatile unsigned char I_MAX = 0;
  volatile unsigned char MODE = 0;                //0..3 corresponds to Mode A..D
  volatile unsigned char OUTPUT_EN = 0;           //Global variable for the state of the output 
  volatile unsigned char OUTP_SHOULD_EN = 1;      //Global variable for higher priorized switching of the output over serial commands. An unwanted switching on after reconnecting the interlock is prohibited, but not switching off when the interlock opens.


  String COMMAND_STR;                    //global variable for storing the serial command
  
  volatile long BAUDRATE = 9600;
  volatile unsigned int TEMPERATURE;
  volatile bool INTERLOCK_TRIPPED = false;
  volatile int ERROR_QUEUE [ERROR_QUEUE_LENGTH] = {0};
  
/***Funktion declarations***/

void serialStringReceive();
int read_temp_sensor();
void change_mode();
void read_commandstring(String new_command);
int read_dip_sw();
void set_ducy(unsigned char new_ducy);
void set_current(unsigned char new_curr);
void add_error_to_queue(int error_id);
void read_remove_error_from_queue();
bool error_already_set(int error_id);
void print_error(int error_id);
void output_enable();
void output_disable();



/***Main program***/

void setup() {
  //Direct register manipulation of ATMEGA328P for changing the PWM frequency
  TCCR2B = 0b00000010;                            //Prescaler 8 for timer 2
  TCCR2A = 0b10100011;                            // Fast PWM Mode, gives ca. 7.8kHz PWM frequency with prescaler

  wdt_disable;                                    //Watchdog Timer is disabled, will only be armed for a software reset

  pinMode(13, OUTPUT);
  pinMode(AIN_INT, OUTPUT);
  set_current(0);
  pinMode(STROBE, OUTPUT);
  set_ducy(255);
  digitalWrite(AIN_INT_EN, LOW);                  //deactivate all inputs for now
  digitalWrite(AIN_EXT_EN, LOW);
  pinMode(AIN_EXT_EN, OUTPUT);   
  pinMode(AIN_INT_EN, OUTPUT);                   
  OUTPUT_EN = 0;

  pinMode(DIP1, INPUT_PULLUP);      
  pinMode(DIP2, INPUT_PULLUP); 
  pinMode(0, INPUT_PULLUP);                       //RX and TX with Pullup-Widerstand to prevent glitches when switching the baud rate
  pinMode(1, INPUT_PULLUP);

  digitalWrite(PC6, HIGH);
  pinMode(PC6, OUTPUT);

  Serial.begin(BAUDRATE);                         //Default 9600 Baud, 8N1
  COMMAND_STR.reserve(50);
}

void loop() {
  char temp_char;
  int analog_val;
  static int mode_new, mode_old = -1;             //Variable for position of DIP switches, does not directly match the operating mode              
  static bool comm_complete = false; 
  static bool local_op = true;

  TEMPERATURE = read_temp_sensor();
  if(TEMPERATURE > EEPROM.read(TEMPERATURE_LIM_ADDR) && !error_already_set(929)){             
    add_error_to_queue(929);                                            //Command usually used for plug-in modules will be used here
  }

  INTERLOCK_TRIPPED = !digitalRead(INTERLOCK);
  if(INTERLOCK_TRIPPED){                                          //At disconnected Interlock (Pin LOW): Set Flag, deactivate output and give out error
    output_disable();
    if(!error_already_set(932)){
      add_error_to_queue(932);
    }
    if(mode_old == 3){
      OUTP_SHOULD_EN = 0;                                         //When operating remotely: No automatic switching on again.
    }
  }
  else if (!INTERLOCK_TRIPPED && OUTPUT_EN == 0 && OUTP_SHOULD_EN == 1){                             //When Interlock is closed again and Flag is still set: clear Flag and activate output  
    output_enable();
    INTERLOCK_TRIPPED =  false;
  }
  
  mode_new = read_dip_sw();                                             //DIP switch is also read when interlock is disconnected, in remote operation modes are always switched, in local operation modes are only switched when interlock is connected
    
  if(mode_new == 3 && mode_new != mode_old){                            //Dip switches = 0b11 -> Remote Operation
    MODE = 0;                                                           //Default is Mode A
    mode_old = mode_new;
    local_op = false;
    change_mode();
  }

  else if(mode_new != mode_old){
    MODE = mode_new;                                                          
    mode_old = mode_new;
    local_op = true;  
    switch (MODE){
      case 0:                                                         //Dip switches = 0b00 -> Local Operation, Mode A, Duty Cycle 100%
        DUTYCYCLE_INV = 0;
        set_ducy(DUTYCYCLE_INV);
        break;
          
      case 1:                                                         //Dip switches = 0b01 -> Local Operation, Mode B, I_MAX 100% 
        I_MAX = 255;
        set_current(I_MAX);
        break;
          
      case 2:                                                         //Dip switches = 0b10 -> Local Operation, Mode C, Duty Cycle 100%, I_MAX 100% 
        I_MAX = 255;
        DUTYCYCLE_INV = 0;
        set_current(I_MAX);
        set_ducy(DUTYCYCLE_INV); 
        break;
    }
  }

  if(comm_complete == false && !local_op && Serial.available()){      //Reiceiving routing for string, no own function to keep free processor time
    temp_char = Serial.read();

    if(temp_char == '\n'){
      comm_complete = true;
      Serial.flush();
    }

    else{
      COMMAND_STR.concat(temp_char);
    }
  }                                                     
  

  if(comm_complete && !local_op){
    read_commandstring(COMMAND_STR);
    comm_complete = false;
    COMMAND_STR = "";   
  }

  switch (MODE){
    case 0:                                             //Mode A: Read, at Uin > switch of output at 3,3V+10%
      if(analogRead(AIN_EXT) > 743){
        output_disable();
        if(!error_already_set(144)){
          add_error_to_queue(144);                      // Input overpower (error is normally used for VNAs)
        }
        else if(local_op && !INTERLOCK_TRIPPED){
          output_enable();                              //In local operation, output is automatically switched on again (only if interlock is connected)
        }
      }

      break;

    case 1:                                             //Mode B: Read ADC, set dutycycle correspondingly
      analog_val = analogRead(AIN_EXT);
      analog_val = constrain (analog_val, 0, 675);
      analog_val = map(analog_val, 0, 675, 255, 0);     //Mapping function inverts dutycycle
        set_ducy(analog_val);
      break;

    default:
      ;
  } 
}


/***ISR***/

void change_onoff_ISR (){                 //ISR for MODE 2 (Mode C, On/Off via interrupt)
  if(MODE == 2){
    if(PIND & 0b00000100){               //Set cutycycle to current target value at high level
      set_ducy(DUTYCYCLE_INV);
    }

    else{
      set_ducy(255);           //Set dutycycle to 0 at low level (inverted)
    }
  }  
}


/***function implementations***/

int read_temp_sensor(){                                   // Attention: scaled to eighths of an degree
  long temp_val = analogRead(A6);
  temp_val = (-12800 + temp_val*125) >> 5;                //Scaling of ADC value to temperature, 500mV voltage (ADC: 12800/125 bzw 102,4) correspond to 0°C, 10mV/°C correspond to 2,048 (= (2^8)/125) LSB/°C 

  return (int)temp_val;
}

void change_mode(){
  
  noInterrupts();
  detachInterrupt(digitalPinToInterrupt(AIN_EXT_INTERR));
  I_MAX = 0;
  DUTYCYCLE_INV = 255;
  set_current(I_MAX);                                     //Set dutycycle and I_Max to 0 during change, STROBE is inverted
  set_ducy(DUTYCYCLE_INV);
  
  switch (MODE) { 

    case 0:                                               // Mode A: Analog input is fed directly to current controller, dutycycleis set by software, input is monitored via ADC, output is switched off over 3.3V
      digitalWrite(AIN_INT_EN, LOW);      
      digitalWrite(AIN_EXT_EN, HIGH);
    break;
    
    case 1:                                               //Mode B: Analog input is read by ADC and controls dutycycle, I_MAX is set by software
      digitalWrite(AIN_INT_EN, HIGH);
      digitalWrite(AIN_EXT_EN, LOW);
    break;
    
    case 2:                                               //Mode C: On/Off via interrupt, dutycycle and I_MAX are set by software
      attachInterrupt(digitalPinToInterrupt(AIN_EXT_INTERR), change_onoff_ISR, CHANGE);
      digitalWrite(AIN_INT_EN, HIGH);
      digitalWrite(AIN_EXT_EN, LOW);
    break;

    case 3:                                               //Mode D: completely digital control, ON/OFF, dutycycle and I_MAX are set by software
      digitalWrite(AIN_INT_EN, HIGH);     
      digitalWrite(AIN_EXT_EN, LOW);
    break;

    default:
    ;
    break;    
  }
  interrupts();
}



void read_commandstring(String new_command){
  int i = 0, j = 0, i_max;
  long baudrate_temp;
  int pos_of_colon[COLON_QUEUE_LENGTH];                   //Array with positions of ':', aka the seperator of subcommands
  bool valid_command = false;

  String rootcommand;                                     //String for Root-Level of SCPI Command
  String subcommand_1;                                    //String for remaining levels, can be furtherly resolved when needed (see further below)
  String subcommand_2;
  String subcommand_3;
  String argument = "      ";                             //Argument of the command
  
  new_command.toUpperCase();                              //creating case insensitivity
  i_max = new_command.length();
  
  for(i = 0; i <= i_max; i++){                            //Seacrh for ':' and save positions
    if (new_command.charAt(i) == ':'){
       pos_of_colon[j] = i;
       j++;
    }
  }
  
  for( ;j < COLON_QUEUE_LENGTH; j++){                     //Fill up rest of the array with i_max, so that commands with less subcommands are recognized, j is already intialized
    pos_of_colon[j] = i_max;
  }
  
  
  rootcommand =  new_command.substring(0, pos_of_colon[0]);
  subcommand_1 = new_command.substring((pos_of_colon[0]+1), pos_of_colon[1]);   //Here, further resolving of subcommands can be implemented if the command set is changed
  subcommand_2 = new_command.substring((pos_of_colon[1]+1), pos_of_colon[2]);
  subcommand_3 = new_command.substring((pos_of_colon[2]+1), pos_of_colon[3]);

  if(rootcommand.startsWith("*IDN?")){
    Serial.print("IR-LED Illuminator, Hardware V1.1, Firmware V1.1, 2021-12-06");
    Serial.print("\n");
    valid_command = true;
  }

  else if(rootcommand.startsWith("*RST")){
    wdt_enable(WDTO_2S);                                  //aktivate Watchdog timer (2 sec, longer than powerup of bootloader) for Reset
    while(1);                                             //ininite loop to trigger WDT
  }

  else if (rootcommand.startsWith("*CLS")){               //*CLS: delete error queue
    for(i = 0; i < ERROR_QUEUE_LENGTH; i++){
      ERROR_QUEUE [i] = 0;
      valid_command = true;
    }
  }

  else if(rootcommand.startsWith("OUTP")){
    if(subcommand_1.startsWith("DUT")){
      if(MODE == 1){                                        //In Mode B, the dutydycle can't be set by SW, give out error -221, "Conflicting Setting"
        add_error_to_queue(-221);  
        valid_command = true;
      }

      else{
        if(subcommand_1.endsWith("?")){                     //Command "DUT?"
          Serial.print(255 - DUTYCYCLE_INV);
          Serial.print("\n");
          valid_command = true;
        }
  
        else{                                               //Command "DUT <>"
          i_max = subcommand_1.length();
          i = subcommand_1.indexOf(" ");                    //Space -> separator of command and argument
          argument = subcommand_1.substring((i+1),i_max);
          j = argument.toInt();
          valid_command = true;
          if(j >= 0 && j < 256){
            if((j==0) == (subcommand_1.endsWith(" 0"))){    //Check if the argument is a number: .toInt() gives back 0 when fed with letters, so the valid case argument = 0 has to be dealt with seperately. Check .endsWith(" 0") only works, when a single zero and not "00" or similar is sent                     
              DUTYCYCLE_INV = 255-j;
              if(!INTERLOCK_TRIPPED){
                set_ducy(DUTYCYCLE_INV);                    // Dutycycle is inverted
              }
            }
          }
  
          else{
            add_error_to_queue(-222);                       //SCPI Error -222: Data out of range
          }
        }
      } 
    }
      
    else if(subcommand_1.startsWith("CUR")){
      if(MODE == 0){                                        //In Mode A, current can't be set by software, give out Error -221, Conflicting Settings
        add_error_to_queue(-221);
        valid_command =  true;
      }

      else{
        if(subcommand_1.endsWith("?")){                     //command "CUR?"
          Serial.print(I_MAX);                              //Not to be confused: I_MAX for max. current, i_max string length
          Serial.print("\n");
          valid_command = true;
        }
  
        else{                                               //command "CUR <>"
          i_max = subcommand_1.length();
          i = subcommand_1.indexOf(" ");                    //Space -> separator of command and argument
          argument = subcommand_1.substring((i+1),i_max);
          j = argument.toInt();
          valid_command = true;
          if(j >= 0 && j < 256){                            
            if((j==0) == (subcommand_1.endsWith(" 0"))){    //Check if the argument is a number: .toInt() gives back 0 when fed with letters, so the valid case argument = 0 has to be dealt with seperately. Check .endsWith(" 0") only works, when a single zero and not "00" or similar is sent    
              I_MAX =  j;
              set_current(I_MAX);
            }
          }
  
          else{
            add_error_to_queue(-222);                       //SCPI Error -222: Data out of range
          }
        }  
      }
    }

    else if(subcommand_1.startsWith("MOD")){
      valid_command =  true;
      if(subcommand_1.endsWith("?")){                     //command "MOD?"
        Serial.print(char(MODE + 'A'));
        Serial.print("\n");
      }

      else{                                               //command "MOD <>"
        i_max = subcommand_1.length();
        i = subcommand_1.indexOf(" ");                    //Space -> separator of command and Argument
        argument = subcommand_1.substring((i+1),i_max);

        if(argument == "A"){
          MODE = 0;
        }
        
        else if(argument == "B"){
          MODE = 1;
        }
        
        else if(argument == "C"){
          MODE = 2;
        }
        
        else if(argument == "D"){
          MODE = 3;
        }
        
        else{                            
          add_error_to_queue(-222);                       //SCPI Error -222: Data out of range
        } 

        change_mode();
      }      
    }

    else if(subcommand_1.startsWith("EN")){  
      if(subcommand_1.endsWith("?")){
        Serial.print(OUTPUT_EN);
        Serial.print("\n");
      }

      else if(subcommand_1.endsWith("1")){
        output_enable();
        OUTP_SHOULD_EN = 1;                               //Superordinate parameter to prevent unwanted switching on of the output, see declaration of global variable
      }

      else if(subcommand_1.endsWith("0")){
        output_disable();
        OUTP_SHOULD_EN = 0;    
      }
      valid_command = true;
    }

  }

  else if(rootcommand.startsWith("TEMP")){
    if(rootcommand.endsWith("?")){
      Serial.print(TEMPERATURE);
      Serial.print("\n");
      valid_command = true;
    }

    else if(subcommand_1.startsWith("LIM")){
      if(subcommand_1.endsWith("?")){
        j = (int) EEPROM.read(TEMPERATURE_LIM_ADDR);
        Serial.print(j);
        Serial.print("\n");
        valid_command =  true;
      }

      else{
        i = subcommand_1.indexOf(" ");                  //Space -> separator of command and Argument
        i_max = subcommand_1.length();
        argument = subcommand_1.substring((i+1),i_max);
        j = argument.toInt();
        if(j >= 0 && j < 256){
          if((j==0) == (subcommand_1.endsWith(" 0"))){    //Check if the argument is a number: .toInt() gives back 0 when fed with letters, so the valid case argument = 0 has to be dealt with seperately. Check .endsWith(" 0") only works, when a single zero and not "00" or similar is sent
            EEPROM.update(TEMPERATURE_LIM_ADDR, (byte) j);
            valid_command =  true;
          }
        }

        else{
          add_error_to_queue(-222);                     //SCPI Error -222: Data out of range
        }
      }
    }   
  }


  else if(rootcommand.startsWith("SYST")){
    if(subcommand_1.startsWith("ERR")){                   //Error commands
      if(subcommand_1.endsWith("?")){                     //command SYST:ERR?
        read_remove_error_from_queue();
        valid_command =  true;
        Serial.print("\n");
      }
  
      else if(subcommand_2.startsWith("ALL")){            //command SYST:ERR:ALL?
        read_remove_error_from_queue();                   //Give back "No error" once when no error is present 
        while (ERROR_QUEUE[0] != 0){
          Serial.print(";");                              //Semikolon als as seperator of error messages
          read_remove_error_from_queue();
        }
        valid_command = true;
        Serial.print("\n");
      }
  
      else if(subcommand_2.startsWith("CODE")){           
        if(subcommand_3.startsWith("COUN") && subcommand_3.endsWith("?")){    //command SYST:ERR:CODE:COUN?
          j = 0;
          for(i = 0; i < ERROR_QUEUE_LENGTH; i++){
            if(ERROR_QUEUE[i] !=0){
              j++;
            }
          }
          Serial.print(j);
          Serial.print("\n");
          valid_command = true;
        }
  
        else if(subcommand_3.startsWith("ALL") && subcommand_3.endsWith("?")){  //command SYST:ERR:CODE:ALL?
          j = 0;
          for(i = ERROR_QUEUE_LENGTH -1; i >= 0; i--){      // count backwards due to FIFO
            if(ERROR_QUEUE[i] !=0){
              if(j != 0){
                Serial.print("; ");                         //When loop is iterated more than once, set Semikolon in front of the next error code
              }
              Serial.print(ERROR_QUEUE[i]);
              ERROR_QUEUE[i] = 0;
              j++;
            }
          }
  
          if(j == 0){
            Serial.print(ERROR_QUEUE[0]);                   //Give out 0 when no error is present
          }
          Serial.print("\n");
          valid_command =  true;
        }   
      }
    }

    else if(subcommand_1.startsWith("COMM") && subcommand_2.startsWith("SER")){
      if(subcommand_3.startsWith("BAUD")){
        if(subcommand_3.endsWith("?")){                   //command SYST:COMM:SER:BAUD?
          Serial.print(BAUDRATE);
          Serial.print("\n");
        }

        else{                                             //command SYST:COMM:SER:BAUD <>
          i_max = subcommand_3.length();      
          i = subcommand_3.indexOf(" ");                          //Space -> separator of command and Argument
          argument = subcommand_3.substring(i+1, i_max-2);        //cut off last two dec. places: equal to x % 100, necessary since String.toInt only goes up to ca. 32k
          baudrate_temp = argument.toInt() * 100;
          argument = subcommand_3.substring(i_max-2, i_max );
          baudrate_temp += argument.toInt();
          valid_command = true;
          
          if(baudrate_temp >= 300 && baudrate_temp <= 2000000){
            BAUDRATE = baudrate_temp;  
            Serial.flush();                         
            Serial.begin(BAUDRATE);
          }

          else{
            add_error_to_queue(-222);                       //SCPI Error -222: Data out of range
          }
        }
      }
    }

    else if(subcommand_1.startsWith("INT")){
      if(subcommand_2.startsWith("TRIP") && subcommand_2.endsWith("?")){    //command SYST:INT:TRIP?
        Serial.print(INTERLOCK_TRIPPED);
        Serial.print("\n");
        valid_command = true;
      }
    }    
  }
    
    


  if(valid_command == false){
    add_error_to_queue(-100);                             //Error -100: Command Error
  }
  
}



int read_dip_sw(){
  int mode = 0;
  mode += digitalRead(DIP1);
  mode += (digitalRead(DIP2))<<1;
  
  return mode;
}



void set_ducy(unsigned char new_ducy){                    //Direct manipulation of AVR registers, quicker than Arduino functions
  if(new_ducy == 0){
    PORTD &= 0b11110111;
    TCCR2A &= 0b11001111;                                 //COM2B = 0b00 -> PD3 is normal IO-Port
  }

  else if(new_ducy == 255){
    PORTD |= 0b00001000;
    TCCR2A &= 0b11001111;                                 
  }

  else{
    OCR2B = new_ducy;
    TCCR2A = 0b00100000 | (TCCR2A & 0b11001111);          //COM2B = 0b10 -> PD3 in noninverting PWM Mode
  }
}



void set_current(unsigned char new_curr){
  if(new_curr == 0){
    PORTB &= 0b11110111;
    TCCR2A &= 0b00111111;                                 //COM2A = 0b00 -> PD3 is normal IO-Port
  }

  else if(new_curr == 255){
    PORTB |= 0b00001000;
    TCCR2A &= 0b00111111;                              
  }

  else{
    OCR2A = new_curr;
    TCCR2A = 0b10000000 | (TCCR2A & 0b00111111);          //COM2A = 0b10 -> PD3 in noninverting PWM Mode
  }  
}


bool error_already_set(int error_id){
  bool error_set =  false;
  int i;
  for (i = 0; i < ERROR_QUEUE_LENGTH; i++){
    if(ERROR_QUEUE[i] == error_id){
      error_set = true;
    }
  }
  return error_set;
}


void add_error_to_queue(int error_id){
  char i;
  for(i = (ERROR_QUEUE_LENGTH - 2); i >= 0; i--){         //Move Queue one further
    ERROR_QUEUE[i+1] = ERROR_QUEUE [i];
  }
  if(ERROR_QUEUE[ERROR_QUEUE_LENGTH - 1] != 0 && ERROR_QUEUE[ERROR_QUEUE_LENGTH - 1] != -350){
    ERROR_QUEUE[0] = -350;                                // Error -350: Error Queue Overflow
  }
  else if(ERROR_QUEUE[ERROR_QUEUE_LENGTH - 1] != -350){
    ERROR_QUEUE[0] = error_id;
  }
}


void read_remove_error_from_queue(){
  int error = 0;                             //Move queue back by one
  char i;
  for (i = (ERROR_QUEUE_LENGTH - 1); i >= 0; i--){
    if(ERROR_QUEUE[i] != 0){
      error =  ERROR_QUEUE[i];
      ERROR_QUEUE[i] = 0;
      break;
    }
  }
  print_error(error);
}


void print_error(int error_id){
  Serial.print(error_id);
  switch (error_id){
  case 932:
    Serial.print(", \"Safety Interlock prevents operation\" ");
    break;
    
  case 929:
    Serial.print(", \"Overtemperature\"");
    break;

  case 144:
    Serial.print(", Input overpower");
    break;

  case 0:
    Serial.print(", No error");
    break;

  case -100:
    Serial.print(", Command Error");
    break;

  case -221:
    Serial.print(", Settings Conflict");
    break;

  case -222:
    Serial.print(", Data out of range");
    break;

  case -350:
    Serial.print(", Error Queue Overflow");
    break;

  default:
    Serial.print(", Error not implemented, something went wrong.");
    
  }
}

void output_enable(){
  if(!INTERLOCK_TRIPPED){
    if(MODE == 0){
      digitalWrite(AIN_INT_EN, LOW);      
      digitalWrite(AIN_EXT_EN, HIGH);
    }
  
    else{
      digitalWrite(AIN_INT_EN, HIGH);      
      digitalWrite(AIN_EXT_EN, LOW);
    }
  
    if(MODE == 2){
      attachInterrupt(digitalPinToInterrupt(AIN_EXT_INTERR), change_onoff_ISR, CHANGE);
    }
    set_ducy(DUTYCYCLE_INV);
    OUTPUT_EN = 1;
  }
}

void output_disable(){
  detachInterrupt(digitalPinToInterrupt(AIN_EXT_INTERR));
  digitalWrite(AIN_EXT_EN, LOW);
  digitalWrite(AIN_INT_EN, LOW);
  set_ducy(255);
  OUTPUT_EN = 0;
}
